/*---------------------------------------------------------------------------*\
                          ____  _ _ __ _ __  ___ _ _
                         |_ / || | '_ \ '_ \/ -_) '_|
                         /__|\_, | .__/ .__/\___|_|
                             |__/|_|  |_|
\*---------------------------------------------------------------------------*/
#ifndef UTILS_COLORS_H_
#define UTILS_COLORS_H_

#include <zypp-tui/utils/colors.h>

using ztui::do_colors;
using ztui::do_ttyout;
using ztui::ColorContext;
using ztui::customColorCtor;
using ztui::CCString;

using ztui::DEFAULTString;

using ztui::MSG_STATUSString;
using ztui::MSG_ERRORString;
using ztui::MSG_WARNINGString;

using ztui::POSITIVEString;
using ztui::CHANGEString;
using ztui::NEGATIVEString;

using ztui::HIGHLIGHTString;
using ztui::LOWLIGHTString;

using ztui::operator<<;

#endif // UTILS_COLORS_H_
