#ifndef ZYPPER_COMMANDS_REPOS_H_INCLUDED
#define ZYPPER_COMMANDS_REPOS_H_INCLUDED

#include "repos/list.h"
#include "repos/add.h"
#include "repos/remove.h"
#include "repos/rename.h"
#include "repos/modify.h"
#include "repos/refresh.h"
#include "repos/clean.h"

#endif
