#ifndef ZYPPER_COMMANDS_UTILS_H_INCLUDED
#define ZYPPER_COMMANDS_UTILS_H_INCLUDED

#include "utils/targetos.h"
#include "utils/versioncmp.h"
#include "utils/licenses.h"
#include "utils/download.h"
#include "utils/source-download.h"
#include "utils/system-architecture.h"
#include "utils/purge-kernels.h"

#endif
