#ifndef ZYPPER_COMMANDS_SERVICES_H_INCLUDED
#define ZYPPER_COMMANDS_SERVICES_H_INCLUDED

#include "services/list.h"
#include "services/refresh.h"
#include "services/add.h"
#include "services/modify.h"
#include "services/remove.h"

#endif
