#ifndef ZYPPER_COMMANDS_QUERY_H_INCLUDED
#define ZYPPER_COMMANDS_QUERY_H_INCLUDED

#include "query/info.h"
#include "query/packages.h"
#include "query/patches.h"
#include "query/patterns.h"
#include "query/products.h"

#endif
